@@ -17,16 +17,18 @@
 Exception related utilities.
 """
 
+import functools
+import io
 import logging
 import os
 import sys
 import time
 import traceback
 
-import six
 
-from oslo_utils._i18n import _LE
+from oslo_utils import encodeutils
 from oslo_utils import reflection
+from oslo_utils import timeutils
 
 
 class CausedByException(Exception):
@@ -43,9 +45,11 @@ class CausedByException(Exception):
                   should itself be an exception instance, this is useful for
                   creating a chain of exceptions for versions of python where
                   this is not yet implemented/supported natively.
+
+    .. versionadded:: 2.4
     """
     def __init__(self, message, cause=None):
-        super(CausedByException, self).__init__(message)
+        super().__init__(message)
         self.cause = cause
 
     def __bytes__(self):
@@ -65,7 +69,7 @@ class CausedByException(Exception):
         if indent < 0:
             raise ValueError("Provided 'indent' must be greater than"
                              " or equal to zero instead of %s" % indent)
-        buf = six.StringIO()
+        buf = io.StringIO()
         if show_root_class:
             buf.write(reflection.get_class_name(self, fully_qualified=False))
             buf.write(": ")
@@ -92,7 +96,6 @@ class CausedByException(Exception):
                     buf.write(line)
                     if i + 1 != len(lines):
                         buf.write(os.linesep)
-            if not isinstance(next_up, CausedByException):
                 # Don't go deeper into non-caused-by exceptions... as we
                 # don't know if there exception 'cause' attributes are even
                 # useable objects...
@@ -124,6 +127,8 @@ def raise_with_cause(exc_cls, message, *args, **kwargs):
                  exceptions constructor.
     :param kwargs: any additional keyword arguments to pass to the
                    exceptions constructor.
+
+    .. versionadded:: 1.6
     """
     if 'cause' not in kwargs:
         exc_type, exc, exc_tb = sys.exc_info()
@@ -133,11 +138,11 @@ def raise_with_cause(exc_cls, message, *args, **kwargs):
         finally:
             # Leave no references around (especially with regards to
             # tracebacks and any variables that it retains internally).
-            del(exc_type, exc, exc_tb)
-    six.raise_from(exc_cls(message, *args, **kwargs), kwargs.get('cause'))
+            del (exc_type, exc, exc_tb)
+    raise exc_cls(message, *args, **kwargs) from kwargs.get('cause')
 
 
-class save_and_reraise_exception(object):
+class save_and_reraise_exception:
     """Save current exception, run some code and then re-raise.
 
     In some cases the exception context can be cleared, resulting in None
@@ -172,55 +177,185 @@ class save_and_reraise_exception(object):
               [if statements to determine whether to raise a new exception]
               # Not raising a new exception, so reraise
               ctxt.reraise = True
+
+    .. versionchanged:: 1.4
+       Added *logger* optional parameter.
     """
     def __init__(self, reraise=True, logger=None):
         self.reraise = reraise
         if logger is None:
             logger = logging.getLogger()
         self.logger = logger
+        self.type_, self.value, self.tb = (None, None, None)
+
+    def force_reraise(self):
+        if self.type_ is None and self.value is None:
+            raise RuntimeError("There is no (currently) captured exception"
+                               " to force the reraising of")
+        try:
+            if self.value is None:
+                self.value = self.type_()
+            if self.value.__traceback__ is not self.tb:
+                raise self.value.with_traceback(self.tb)
+            raise self.value
+        finally:
+            self.value = None
+            self.tb = None
+
+    def capture(self, check=True):
+        (type_, value, tb) = sys.exc_info()
+        if check and type_ is None and value is None:
+            raise RuntimeError("There is no active exception to capture")
+        self.type_, self.value, self.tb = (type_, value, tb)
+        return self
 
     def __enter__(self):
-        self.type_, self.value, self.tb, = sys.exc_info()
-        return self
+        # TODO(harlowja): perhaps someday in the future turn check here
+        # to true, because that is likely the desired intention, and doing
+        # so ensures that people are actually using this correctly.
+        return self.capture(check=False)
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         if exc_type is not None:
             if self.reraise:
-                self.logger.error(_LE('Original exception being dropped: %s'),
+                self.logger.error('Original exception being dropped: %s',
                                   traceback.format_exception(self.type_,
                                                              self.value,
                                                              self.tb))
             return False
         if self.reraise:
-            six.reraise(self.type_, self.value, self.tb)
+            self.force_reraise()
 
 
-def forever_retry_uncaught_exceptions(infunc):
-    def inner_func(*args, **kwargs):
-        last_log_time = 0
-        last_exc_message = None
-        exc_count = 0
-        while True:
-            try:
-                return infunc(*args, **kwargs)
-            except Exception as exc:
-                this_exc_message = six.u(str(exc))
-                if this_exc_message == last_exc_message:
-                    exc_count += 1
+def forever_retry_uncaught_exceptions(*args, **kwargs):
+    """Decorates provided function with infinite retry behavior.
+
+    The function retry delay is **always** one second unless
+    keyword argument ``retry_delay`` is passed that defines a value different
+    than 1.0 (less than zero values are automatically changed to be 0.0).
+
+    If repeated exceptions with the same message occur, logging will only
+    output/get triggered for those equivalent messages every 60.0
+    seconds, this can be altered by keyword argument ``same_log_delay`` to
+    be a value different than 60.0 seconds (exceptions that change the
+    message are always logged no matter what this delay is set to). As in
+    the ``retry_delay`` case if this is less than zero, it will be
+    automatically changed to be 0.0.
+    """
+
+    def decorator(infunc):
+        retry_delay = max(0.0, float(kwargs.get('retry_delay', 1.0)))
+        same_log_delay = max(0.0, float(kwargs.get('same_log_delay', 60.0)))
+
+        @functools.wraps(infunc)
+        def wrapper(*args, **kwargs):
+            last_exc_message = None
+            same_failure_count = 0
+            watch = timeutils.StopWatch(duration=same_log_delay)
+            while True:
+                try:
+                    return infunc(*args, **kwargs)
+                except Exception as exc:
+                    this_exc_message = encodeutils.exception_to_unicode(exc)
+                    if this_exc_message == last_exc_message:
+                        same_failure_count += 1
+                    else:
+                        same_failure_count = 1
+                    if this_exc_message != last_exc_message or watch.expired():
+                        # The watch has expired or the exception message
+                        # changed, so time to log it again...
+                        logging.exception(
+                            'Unexpected exception occurred %d time(s)... '
+                            'retrying.' % same_failure_count)
+                        if not watch.has_started():
+                            watch.start()
+                        else:
+                            watch.restart()
+                        same_failure_count = 0
+                        last_exc_message = this_exc_message
+                    time.sleep(retry_delay)
+        return wrapper
+
+    # This is needed to handle when the decorator has args or the decorator
+    # doesn't have args, python is rather weird here...
+    if kwargs or not args:
+        return decorator
+    else:
+        if len(args) == 1:
+            return decorator(args[0])
+        else:
+            return decorator
+
+
+class exception_filter:
+    """A context manager that prevents some exceptions from being raised.
+
+    Use this class as a decorator for a function that returns whether a given
+    exception should be ignored, in cases where complex logic beyond subclass
+    matching is required. e.g.
+
+    >>> @exception_filter
+    >>> def ignore_test_assertions(ex):
+    ...     return isinstance(ex, AssertionError) and 'test' in str(ex)
+
+    The filter matching function can then be used as a context manager:
+
+    >>> with ignore_test_assertions:
+    ...     assert False, 'This is a test'
+
+    or called directly:
+
+    >>> try:
+    ...     assert False, 'This is a test'
+    ... except Exception as ex:
+    ...     ignore_test_assertions(ex)
+
+    Any non-matching exception will be re-raised. When the filter is used as a
+    context manager, the traceback for re-raised exceptions is always
+    preserved. When the filter is called as a function, the traceback is
+    preserved provided that no other exceptions have been raised in the
+    intervening time. The context manager method is preferred for this reason
+    except in cases where the ignored exception affects control flow.
+    """
+
+    def __init__(self, should_ignore_ex):
+        self._should_ignore_ex = should_ignore_ex
+
+        if all(hasattr(should_ignore_ex, a)
+               for a in functools.WRAPPER_ASSIGNMENTS):
+            functools.update_wrapper(self, should_ignore_ex)
+
+    def __get__(self, obj, owner):
+        return type(self)(self._should_ignore_ex.__get__(obj, owner))
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        if exc_val is not None:
+            return self._should_ignore_ex(exc_val)
+
+    def __call__(self, ex):
+        """Re-raise any exception value not being filtered out.
+
+        If the exception was the last to be raised, it will be re-raised with
+        its original traceback.
+        """
+        exc_type, exc_val, traceback = sys.exc_info()
+
+        try:
+            if not self._should_ignore_ex(ex):
+                if exc_val is ex:
+                    try:
+                        if exc_val is None:
+                            exc_val = exc_type()
+                        if exc_val.__traceback__ is not traceback:
+                            raise exc_val.with_traceback(traceback)
+                        raise exc_val
+                    finally:
+                        exc_val = None
+                        traceback = None
                 else:
-                    exc_count = 1
-                # Do not log any more frequently than once a minute unless
-                # the exception message changes
-                cur_time = int(time.time())
-                if (cur_time - last_log_time > 60 or
-                        this_exc_message != last_exc_message):
-                    logging.exception(
-                        _LE('Unexpected exception occurred %d time(s)... '
-                            'retrying.') % exc_count)
-                    last_log_time = cur_time
-                    last_exc_message = this_exc_message
-                    exc_count = 0
-                # This should be a very rare event. In case it isn't, do
-                # a sleep.
-                time.sleep(1)
-    return inner_func
+                    raise ex
+        finally:
+            del exc_type, exc_val, traceback
